/* Instructable: Vehicular Combat Game: With sound effects and flashing lights!
    By JD_K
    Link to Instructable: https://www.instructables.com/Vehicular-Combat-Game-With-Sound-Effects-and-Flash/ 
    Link to Youtube demo: https://youtu.be/X_o3tAEoQvo
    Oct 2022
*/

// Setting up variable for pins and measurements in the sketch
int IRpinF = A0;               // IR photodiode sensors on analog pin A0 for FRONT
int IRpinR = A1;               // IR photodiode sensors on analog pin A1 for RIGHT
int IRpinL = A2;               // IR photodiode sensors on analog pin A2 for LEFT
int IRpinB = A3;               // IR photodiode sensors on analog pin A3 for BACK
int IRgun = 12;                // IR emitting LED to use as the "gun" on digital pin 12
int gunLight = 11;             // pin 11 for an LED to light up to show what the IRgun is doing - this LED is behind the the IR emitting LED in the translucent gun barrel: light up when shooting (gunTime), dim when reloading (reloadTime)
byte gunToggle = 0;            // simple variable to keep track of what the gun is doing; 1 = shooting; 0 = inactive
int ambientIRF;                // variable to store the ambient IR for the FRONT. Read upon starting and used in the game to compare for determining if the vehicle is hit
int ambientIRR;                // for the Right side
int ambientIRL;                 // for the Left side
int ambientIRB;                 // for the Back
int IRreadF, IRreadR, IRreadL, IRreadB;                   // variable for the temporary recording of IR readings, if significantly lower than ambientIR then the vehicle is hit
int trueVcc;                  // for determining the true value of Vcc for analogRead() measurements
int IRsensitivity = 20;       // to set the sensitivity of the IR sensors (difference between the IRread variables and ambientIRx variables), the lower the more sensitive; try 20 to start
int HitCounter = 0;           // to track the number of hits
long prevMillis = 0;          // used to time how long ago the gun was fired to determine the duration of IR gun activation (gunTime) and and how long until it is ready (reloadTime)
long currentMillis;           // used for timing of things, other time variables will be compared to this
long IRMillis;                // to use to time how long between IR readings for the IRread
int gunTime = 2000;           // Duration, in milliseconds, of IR gun activation
int reloadTime = 3000;        // Duration, in milliseconds, until the IR gun is ready (note: the gunLight will then be off)

//LD293 connections for Left Motor: Enable1 pin of L293D which is M1 PWM is to be pinned to Arduino D5, M1 direction to D4, OUT1 to the motor, 0V to GND, connect both 0V pins together, OUT2 to motor, IN2 of L293D to Arduino D7, and finally +Vmotor pin of L293D to the motor power supply
#define ENA 5 //PWM pin
#define IN1 4
#define IN2 7
//LD293 connections for Right Motor: Enable2 pin of L293D which is M2 PWM is to be pinned to Arduino D6, M2 direction to D9, OUT1 to the motor, 0V to GND, connect both 0V pins together, Out2 to the motor, IN2 of L293D to Arduino D8, and finally +V pin of L293D to the Arduino 5V
#define ENB 6 //PWM pin
#define IN3 9
#define IN4 8
unsigned char carSpeed = 255; //150 is not enough to get the car to turn.255 would be maximum,

#define LED 13 // pin D13 to the headlights: the controller can turn on/off, but these will also flash when hit, and will flash continuously when Game Over
bool LEDstate = LOW; // to toggle to turn the LED on/off

/*  Note: Hall sensors are polarized, one side will detect a north pole of a magnet but not a south pole
    Hall sensor's left pin when facing you (text toward you): Vcc to 5V
    Centre Pin is the sensor's GND - to ARduino GND
    Sensor's right pin is the data pin, which is connected to the sensor's Vcc pin via a 10 kOhm resistor, and the sensor's data pin to Arduino D2
*/
const byte interruptPin = 2;


/*HR-06 Bluetooth receiver and an android App on a phone/tablet to control, remember the Bluetooth Tx and Rx must be DISCONNECTED when uploading code to the Arduino
   Pinout as per BT receiver instructions:
   Rx of HR-06 needs a voltage divider to limit input dgital signal to 3.3V: parallel connection of Rx of the HR-06 via a 1kOhm resistor to Tx of Arduino, and Rx of the HR-06 via a 2kOhm resistor to GND;
   Tx of the HR-06 to Rx of Arduino;
   GND to GND
   Vcc of HR-06 to Arduino 5V
*/
char getstr; //the variable read from the HR-06 and will be used to determine which function to run


// MP3 player
#include <SoftwareSerial.h>
#include <DFPlayer_Mini_Mp3.h>
SoftwareSerial mySerial(A4, A5); //declares the pins on the Arduino which the Dfplayer mini is connected
byte statePin = 10; // added Aug 3 2022: need to declare a pin to see if the MP3 player is busy, in which case that pin would be LOW



// ******************* These Functions are called when the bluetooth HR-06 recieves a command code in the Loop ******************
void forward() {
  analogWrite(ENA, carSpeed);
  analogWrite(ENB, carSpeed);
  digitalWrite(IN1, HIGH);
  digitalWrite(IN2, LOW);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4, HIGH);
}

void back() {
  analogWrite(ENA, carSpeed);
  analogWrite(ENB, carSpeed);
  digitalWrite(IN1, LOW);
  digitalWrite(IN2, HIGH);
  digitalWrite(IN3, HIGH);
  digitalWrite(IN4, LOW);
}

void left() {
  analogWrite(ENA, carSpeed);
  analogWrite(ENB, carSpeed);
  digitalWrite(IN1, LOW);
  digitalWrite(IN2, HIGH);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4, HIGH);
  delay(200);
  digitalWrite(IN2, LOW);
  digitalWrite(IN4, LOW);
}

void right() {
  analogWrite(ENA, carSpeed);
  analogWrite(ENB, carSpeed);
  digitalWrite(IN1, HIGH);
  digitalWrite(IN2, LOW);
  digitalWrite(IN3, HIGH);
  digitalWrite(IN4, LOW);
  delay(200);
  digitalWrite(IN1, LOW);
  digitalWrite(IN3, LOW);
}

void stop() {
  digitalWrite(ENA, LOW);
  digitalWrite(ENB, LOW);
}

void LEDtoggle() { //Turns the LED on/off
  LEDstate = !LEDstate;
  digitalWrite(LED, LEDstate);
}
void Shoot() {
  if (gunToggle == 0) {
    analogWrite(IRgun, 250);
    analogWrite(gunLight, 250);
    prevMillis = millis();
    mp3_next ();
    gunToggle = 1;
  }
}

// The interrupt function which is attached when the car gets hit; if the Hall effect sensor is subsequently triggered this functiion is called to reset the counter back to zero
void ResetHitCounter() {
  detachInterrupt(0); //turn the interrupt off, so this interrupt function cannot be called until it is attached to pin2 again (in the subroutine when an IR sensor is hit)
  HitCounter = 0;
}

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA, ADSC));
  result = ADCL;
  result |= ADCH << 8;
  //  result = 1126400L / result; // Calculate Vcc (in mV); 1126400 = 1.1*1024*1000
  result = 1126.4 / result; // Because I want the corrected 10bi ADC value (0-1023)
  return result;
}

void calibrateIR()
{
  trueVcc = readVcc() ;
  ambientIRF = (analogRead(IRpinF) * trueVcc / 5);
  ambientIRR = (analogRead(IRpinR) * trueVcc / 5);
  ambientIRL = (analogRead(IRpinL) * trueVcc / 5);
  ambientIRB = (analogRead(IRpinB) * trueVcc / 5);
  // return ambientIRF, ambientIRR, ambientIRL, ambientIRB;
}


// ****************************** Setup *************************************

void setup() {
  Serial.begin(9600); // Start serial communication, to be used by Bluetooth

  // for the mp3 player:
  mySerial.begin (9600); //SoftwareSerial using the A4 and A5 pins
  mp3_set_serial (mySerial);    //set softwareSerial for DFPlayer-mini mp3 module
  delay(1);                     // delay 1ms to set volume
  mp3_set_volume (25);          // value 0~30 (27-30 was too much distortion)
  pinMode(statePin, INPUT); // to see if the mp3 player is busy

  // for the LEDs:
  pinMode(LED, OUTPUT);
  pinMode(IRgun, OUTPUT);
  pinMode(gunLight, OUTPUT);

  // for the LD293 motor driver
  pinMode(IN1, OUTPUT);
  pinMode(IN2, OUTPUT);
  pinMode(IN3, OUTPUT);
  pinMode(IN4, OUTPUT);
  pinMode(ENA, OUTPUT);
  pinMode(ENB, OUTPUT);
  stop(); // making sure car isn't moving

  pinMode(interruptPin, INPUT_PULLUP); //Setting the pin for the Hall effect sensor


  // Get an ambient IR readings
  calibrateIR();
}

// **************************** the Loop *************************************
void loop() {

  getstr = Serial.read(); //reads the bluetooth module and records as getstr
  switch (getstr) {
    case 'f': forward(); break;
    case 'b': back();   break;
    case 'l': left();   break;
    case 'r': right();  break;
    case 's': stop();   break;
    case 'a': LEDtoggle(); break;
    case 'x': Shoot(); break;
    default:  break;
  }


  currentMillis = millis() ; //Reset currentMillis to current time to compare to the following 'if' statements

  if ((gunToggle == 1) && (currentMillis - prevMillis >= gunTime)) {
    // the gun is shooting, and it has reached the end of the gunTime, so turn off and enter the "reload" time
    digitalWrite(IRgun, LOW);
    analogWrite(gunLight, 50);
    if ((currentMillis - prevMillis) >= (gunTime + reloadTime)) {
      analogWrite(gunLight, 0); // if enough time has passed then turn off the gunLight in the gun to show it is ready again
      prevMillis = millis();
      gunToggle = 0; // now the gun can shoot again
    }
  }

  if (currentMillis - IRMillis >= 100) {
    trueVcc = readVcc(); // correcting for the fluctuating Vcc
    IRreadF = (analogRead(IRpinF) * trueVcc / 5); // Read and print the IR reading every 100 milliseconds;
    IRreadR = (analogRead(IRpinR) * trueVcc / 5);
    IRreadL = (analogRead(IRpinL) * trueVcc / 5);
    IRreadB = (analogRead(IRpinB) * trueVcc / 5);

    IRMillis = millis(); // reset the time of IRMills so that a reading will not be done for another 100 milliseconds

    if ((IRreadF < (ambientIRF - IRsensitivity)) || (IRreadR < (ambientIRR - IRsensitivity)) || (IRreadL < (ambientIRL - IRsensitivity)) || (IRreadB < (ambientIRB - IRsensitivity))) { //the analog signal of the IR sensors will go lower if IR light on them, so if the IRread is lower than ambientIR then it is counted as a hit;
      // if any of the above conditions are true, then the car has been hit
      stop();
      HitCounter = HitCounter + 1;
      if (HitCounter < 3) {
        for (int i = 0; i < 5; i++) { //Flash quickly the number of times it has been hit and repeats this 5 times
          for (int j = 0; j < HitCounter; j++) {
            digitalWrite(LED, HIGH);
            delay (200);
            digitalWrite(LED, LOW);
            delay(300);
          }
          delay(1000);
        }

        digitalWrite(LED, LEDstate); // so if the LED was already on when you were hit, it is back on, otherwise it will be off
        IRreadF = 1024; IRreadR = 1024; IRreadL = 1024; IRreadB = 1024; // so you cannot be high more than once by a single low reading
        IRMillis=(millis() + 5000); //Now the car is invincible for 5 seconds! 
        attachInterrupt(0, ResetHitCounter, CHANGE); // turning on the interrupt so if the Hall sensor changes; the reset function will be called
        mySerial.end();
        mySerial.begin (9600);
        calibrateIR(); // sometimes there are multiple hits in a row with no other car in play, so lets recalibrate now
      } else if (HitCounter > 2) {
        while (HitCounter > 2) { // if 3 Hits, triple flash the LED slowly, and it is Game Over stuck in this while loop until the Arduino is Reset
          for (int j = 0; j < HitCounter; j++) {
            digitalWrite(LED, HIGH);
            delay (200);
            digitalWrite(LED, LOW);
            delay(300);
          }
          delay(2000);
        }
      }
    }
  }
}
